/* $Id: lbunzip2.h,v 1.12.2.5 2008-08-19 15:16:49 lacos Exp $ */

#ifndef LBUNZIP2_H
#  define LBUNZIP2_H

#  define VERSION    "0.03-shift"

/* Maximum size of one full bzip2 stream consisting of one block. */
#  define MX_BZIP2   (((900000lu - 19lu) * 101lu + 99lu) / 100lu + 600lu)

/* 48 bit mask for bzip2 block header and end of stream marker. */
#  define MAGIC_MASK ((uint64_t)0xFFFFlu << 32 | (uint64_t)0xFFFFFFFFlu)

/* 48 bit bzip2 block header. */
#  define MAGIC_HDR  ((uint64_t)0x3141lu << 32 | (uint64_t)0x59265359lu)

/* 48 bit bzip2 end of stream marker. */
#  define MAGIC_EOS  ((uint64_t)0x1772lu << 32 | (uint64_t)0x45385090lu)

/* Number of bytes in CRC. */
#  define NUM_CRC    4u

/* Decompression granularity. */
#  define MX_BUNZIP2 (2lu * 1024lu * 1024lu)

/* At most this many worker threads. */
#  define MX_WORKER  64u

/* Split queue must be able to hold this many blocks for each worker. */
#  define FAC_SPLITQ 4u


struct splitb
{
  struct splitb *next;           /* Next in the same queue. */
  long unsigned serial;          /* Block index as read from stdin. */
  size_t num_fullbytes;          /* Number of bytes loaded. */
  char unsigned bzip2[MX_BZIP2]; /* Stream needing CRC and padding. */
  unsigned incomplete,           /* Last bits. */
      num_freebits;              /* Number of free bits. */
};


struct splitq
{
  struct splitb *lists_base;      /* Kept around for freeing the lists. */

  struct {
    pthread_mutex_t lock;         /* Protects "used". */
    struct splitb *head,          /* Used list. */
        *tail;
    int eof;                      /* Spliter won't emit any more blocks. */
    pthread_cond_t avail_or_eof;  /* Used list not empty or splitter done. */
    long unsigned c_avail_or_eof, /* Check and wait counts for condition. */
        w_avail_or_eof;
  } used;

  struct {
    pthread_mutex_t lock;         /* Protects "free". */
    struct splitb *head,          /* Free list. */
        *tail;
    pthread_cond_t avail;         /* Free list not empty. */
    long unsigned c_avail,        /* Check and wait counts for condition. */
        w_avail;
  } free;
};


struct muxb                          /* Carries a decompressed sub-block. */
{
  long unsigned serial,              /* Block index as read from stdin. */
      subserial;                     /* Sub-block index for the same block. */
  struct splitb *source;             /* Carries source with last subserial. */
  size_t used;                       /* Number of bytes in "decompr" below. */
  char unsigned decompr[MX_BUNZIP2]; /* Decompressed data. */
};


struct muxq
{
  pthread_mutex_t lock;            /* Protects "muxq". */
  struct lacos_rbtree_node *order; /* Orders decompressed sub-blocks. */
  pthread_cond_t avail_or_exit;    /* Next sub-block ready or workers done. */
  long unsigned c_avail_or_exit,   /* Check and wait counts for condition. */
      w_avail_or_exit;
  unsigned num_workers;            /* Number of active workers. */
  long unsigned next_serial,       /* Next sub-block to write out. */
      next_subserial;
};

#endif
